# copyright: Huynh Duong Phuong Vi
# www.sketchupforyou.com
module S4U
    module S4u_connect
        def self.connect()
            model=Sketchup.active_model
            sel=model.selection.to_a
            no_selected=model.active_entities.grep(Sketchup::Edge)-sel
            curver=[]
            edges=[]
            vertex=[]
            sel.each{|e|
                if e && e.valid? && e.is_a?(Sketchup::Edge) && e.length!=0
                    if e.curve
                        curver << e.curve
                    else
                        edges << e
                    end
                    vertex+=e.vertices.select{|v| !v.curve_interior?}
                end
            }
            curver.uniq!
            vers2=[]
            vers1=[]
            vers=vertex.select{|v|(v.edges).length==1}
            vers_sel=vertex.select{|v|(v.edges & sel).length==1}

            if vers.length>1
                vers1=vers[0].edges.collect{|e|
                e.all_connected.grep(Sketchup::Edge).collect{|v| v.vertices}.flatten & vers}.flatten
                vers2=vers-vers1
            elsif vers.length==1
                vers1=vers
                vers2=vers_sel-(vers1[0].edges & sel).collect{|e| e.vertices}.flatten
            elsif vers_sel.length>0
                vers1=(vers_sel[0].edges & sel).collect{|e| e.vertices}.flatten.select{|v| !v.curve_interior?}
                vers2=vers_sel-vers1
            end

            if vers1 && vers1.length>0  && vers2 && vers2.length>0
                vs=vers1.product vers2
                vs.select!{|v| v[0].position.distance(v[1].position)>0.0001}
                vs.sort_by!{|v| v[0].position.distance(v[1].position)}
                v1,v2=vs.find{|v| !v[0].common_edge(v[1])}
            elsif vers1 && vers1.length>1
                (v1,v2=vers1[0],vers1[-1]) if not vers1[0].common_edge vers1[-1]
            elsif vers1 && vers1.length==1
                vers2=vers_sel-vers1
                vs=vers1.product vers2
                vs=vs.sort_by{|v| v[0].position.distance v[1].position}
                v1,v2=vs.find{|v| !v[0].common_edge v[1]}
            elsif vers2 && vers2.length>1
                (v1,v2=vers2[0],vers2[-1]) if not vers2[0].common_edge vers2[-1]
            end
            if v1 && v2 && (not v1.common_edge(v2))
                Sketchup.active_model.start_operation Strings["Connect"],true,false,false
                g=model.active_entities.add_group
                g.entities.add_edges(v1.position,v2.position)
                g.explode
                Sketchup.active_model.commit_operation
                model.selection.add(model.active_entities.grep(Sketchup::Edge) - no_selected)
            end
        end
        def self.point_in_polygon(point,polygon,border=true)
            view=Sketchup.active_model.active_view
            ps=polygon.collect{|e| view.screen_coords(e)}
            return Geom::point_in_polygon_2D(view.screen_coords(point),ps,border)
        end
        def self.draw_circle(point=ORIGIN,vector=Z_AXIS,radius=1,a,vstart,vend)
            c=[]
            numsegs=2*Math::PI/a
            vz=vector
            vx=vstart
            vy=vz.cross vx
            ang=nil
            trans=Geom::Transformation.new(vx,vy,vz,point)
            if vend && (ang=vstart.angle_between(vend)) && ang.abs>=0.00001
                v=vstart.cross vend
                (ang=2*Math::PI-ang) if v.length!=0 && v.samedirection?(vz.reverse)
            end
            if !ang || ang.abs<0.001
                ang=2*Math::PI
            end
            0.step(ang,a).each{|t| c<< ([radius*Math::cos(t),radius*Math::sin(t),0].transform! trans) }
            return c
        end
        def self.get_vertices(vertices)
            vers=[]
            #sel=Sketchup.active_model.selection.to_a
            vers<<vertices[0]
            (1..vertices.length-2).each{|i|
                next if !vertices[i-1] || !vertices[i+1] ||  !vertices[i]
                v1=vertices[i-1].position-vertices[i].position
                v2=vertices[i+1].position-vertices[i].position
                a=v1.angle_between(v2)
                vers<< vertices[i] if a.abs >0.00001 && (a-Math::PI).abs>0.00001
            }
            vers<<vertices[-1]

            return vers
        end
        def self.get_info(c)
            vertices=c.vertices
            #vertices=get_vertices vertices
            reverse=false
            if c.is_a?Sketchup::ArcCurve
                vertices=get_vertices vertices
                center=c.center
                if vertices.length>2
                    a=(vertices[0].position-c.center).normalize.angle_between((vertices[1].position-c.center).normalize)
                    a1=(vertices[2].position-center).normalize.angle_between((vertices[1].position-center).normalize)
                    a=a1 if a1>a
                else
                    a=(vertices[0].position-c.center).normalize.angle_between((vertices[1].position-c.center).normalize)
                end
                return [c.center, c.normal, c.radius,a,vertices]
            end
            p1=vertices[0].position
            p2=vertices[2].position
            p=vertices[1].position
            zaxis=((p2-p).normalize).cross((p1-p).normalize)
            if (p1.distance(p)-p2.distance(p)).abs>0.0001 || zaxis.length<0.00001
                vertices.reverse!
                reverse=true
                p1=vertices[0].position
                p2=vertices[2].position
                p=vertices[1].position
                zaxis=((p2-p).normalize).cross((p1-p).normalize)
                if (p1.distance(p)-p2.distance(p)).abs>0.0001 || zaxis.length<0.00001
                    vertices=vertices[1..-1]
                    vertices<<vertices[0]
                    p1=vertices[0].position
                    p2=vertices[2].position
                    p=vertices[1].position
                    zaxis=((p2-p).normalize).cross((p1-p).normalize)

                end
            end
            point=Geom::Point3d.new(((p1.x+p2.x)/ 2.0),((p1.y+p2.y)/ 2.0),((p1.z+p2.z)/ 2.0))
            vec= (point-p)
            line=[p,vec]
            v1 =( p-p1)
            vec1 = (v1.cross zaxis).normalize
            point1=Geom::Point3d.new(((p1.x+p.x)/ 2.0),((p1.y+p.y)/ 2.0),((p1.z+p.z)/ 2.0))
            line1=[point1,vec1]
            center=Geom::intersect_line_line(line,line1)
            r=p1.distance center
            a=(vertices[0].position-center).normalize.angle_between((vertices[1].position-center).normalize)
            a1=(vertices[2].position-center).normalize.angle_between((vertices[1].position-center).normalize)
            a=a1 if a1>a
            return [center, zaxis, r,a,vertices,reverse]
        end
        def self.get_edges_circle(c)
            center,vector,radius,a,vertices,reverse=get_info c
            vertices_c=c.vertices
            vertices_c.reverse! if reverse
            vector_start=vertices_c[-2].position-center
            vector_end=vertices_c[1].position-center
            ps= draw_circle(center,vector,radius,a,vector_start,vector_end)
            p1=Geom::Point3d.new ps[0]
            ps[0]=vertices_c[-1].position.to_a
            if ((vertices_c[0].position.distance(ps[-1])> vertices_c[0].position.distance(ps[-2])) &&
                (vertices_c[1].position.distance(ps[-1])> vertices_c[0].position.distance(ps[-1]))) ||
                (vertices_c[0].position.distance(ps[-2]) > ps[-1].distance(ps[-2]))
                ps<<vertices_c[0].position.to_a
            else
                ps[-1]=vertices_c[0].position.to_a
            end
            ps.collect!{|p| Geom::Point3d.new(p)}
            return [ps,center,radius]
        end
        def self.curve_is_arc(curve)
            if curve.is_a?Sketchup::Curve
                return [ true,curve.vertices.collect{|e| e.position}]if curve.is_a?Sketchup::ArcCurve
                vs=curve.vertices.select{|v| (v.edges & curve.edges).length==2}
                angles=vs.collect{|v|
                    ((v.edges & curve.edges)[0].other_vertex(v).position-v.position).angle_between(((v.edges& curve.edges)[1].other_vertex(v).position-v.position))}.sort
                if angles.length>1
                    return [((angles[-1]-angles[0]).abs<=0.01 ),curve.vertices.collect{|e| e.position} ]
                elsif angles.length==1 && vs.length==1
                    return [(((vs[0].edges & curve.edges)[0].length-(vs[0].edges & curve.edges)[1].length).abs<=0.0001),curve.vertices.collect{|e| e.position} ]
                end
            else
                eds=curve.each_cons(2).to_a.select{|e| e[0].distance(e[1])>0.0001}
                angles=[]
                (0..eds.length-2).each{|i|
                    angles<< (eds[i][0]-eds[i][1]).angle_between(eds[i+1][1]-eds[i+1][0])
                }
                cur=eds.collect{|e| e[0]}
                cur<<eds[-1][1]
                if angles.length>1
                    return [((angles[-1]-angles[0]).abs<=0.01 ),cur]
                elsif angles.length==1
                    return [((eds[0][0].distance(eds[0][1]) - eds[1][0].distance(eds[1][1]) ).abs<=0.0001),cur]
                end
            end
            return false
        end
        def self.add_edges_extend(g,ps1,closest_points)
            ps1.each{|p|
                # if  p && p[2] && p[2].valid? && (p[2].is_a?Sketchup::Edge) && p[1] && p[1].valid? && (p[1].edges.length==1) && !p[1].edges[0].curve
                #     vector_array=p[0]- p[1].position
                #     if vector_array.length>0
                #         e=g.entities.add_edges p[0].transform(g.transformation.inverse),
                #                                p[1].position.transform(g.transformation.inverse)
                #         if e
                #             pt=e[0].vertices.sort_by{|v| v.position.distance(p[0])}
                #             g.entities.transform_by_vectors [p[1],pt[-1]], [vector_array,vector_array]
                #         end
                #     end
                # els
                if p && p[1] && p[1].valid? && p[2] && p[2].valid? && (p[2].is_a?Sketchup::Edge)
                    g1=g.entities.add_group
                    e=g1.entities.add_edges p[0],p[1].position
                    eds=g1.explode.select{|e| e && e.valid? && e.is_a?(Sketchup::Edge)}
                    eds.each{|e|
                        if p[2]
                            e.layer=p[2].layer
                            e.material=p[2].material
                            e.hidden=true if p[2].hidden?
                            e.soft=true if p[2].soft?
                            e.smooth=true if p[2].smooth?
                        end
                    }
                end
            }
            closest_points.each{|p|
                e=g.entities.add_edges p[0],p[1]
                if p[2] && e && (e=e[0]) && e
                    e.layer=p[2].layer
                    e.material=p[2].material
                    e.hidden=true if p[2].hidden?
                    e.soft=true if p[2].soft?
                    e.smooth=true if p[2].smooth?
                end
            }
            return g
        end
        def self.get_intersect_line_line(edge1,edge2,intersect=true)
            model=Sketchup.active_model
            sel=model.selection.to_a
            points_edge1=[]
            points_edge2=[]
            closest_points=[]
            point=nil
            if intersect
                point = Geom::intersect_line_line(edge1.line,edge2.line)
                point = [point,point] if point
            else
                point=Geom::closest_points(edge1.line,edge2.line)
                closest_points<<point
            end
            if point
                if point_in_polygon(point[1],edge2.vertices.collect{|e| e.position})
                    point2=edge1.vertices.min_by{|v| v.position.distance(point[0])}
                    points_edge1<<[point[0],point2,edge1]
                    point2=edge2.vertices.min_by{|v| v.position.distance(point[0])}
                    points_edge2<<[point[0],point2,edge2]
                else
                    if (edge2.start.edges & sel).length==1 &&
                        point_in_polygon(edge2.start,[point[1],(edge2.start.edges & sel)[0].other_vertex(edge2.start)])
                        points_edge2<<[point[1],edge2.start,edge2]
                        point2=edge1.vertices.min_by{|v| v.position.distance(point[0])}
                        points_edge1<<[point[0],point2,edge1]
                    end
                    if (edge2.end.edges & sel).length==1 &&
                        point_in_polygon(edge2.end,[point[1],(edge2.end.edges & sel)[0].other_vertex(edge2.end)])
                        points_edge2<<[point[1],edge2.end,edge2]
                        point2=edge1.vertices.min_by{|v| v.position.distance(point[0])}
                        points_edge1<<[point[0],point2,edge1]
                    end
                end
            end
            return [points_edge1,points_edge2,closest_points]
        end
        def self.get_intersect_curve_line(curve,edge1,out=false,edges1=[])
            points_edge1=[]
            points_curve=[]
            closest_points=[]
            curve.edges.each{|edge2|
                point= Geom::intersect_line_line(edge1.line,edge2.line)
                if point
                    if point_in_polygon(point,edge2.vertices.collect{|e| e.position})
                        point2=edge1.vertices.min_by{|v| v.position.distance(point)}
                        if point2 && point_in_polygon(point,edge1.vertices.collect{|e| e.position}) &&
                            point2.position.distance(point)>0.1
                            points_edge1<<[point,point2,edge1]
                            points_curve<<edge2
                            # break
                        elsif out
                            if	edge1.curve && (edge1.vertices[0]==edge1.curve.vertices[0]) &&
                                point_in_polygon(edge1.vertices[0].position,[edge1.vertices[1].position,point])
                                points_edge1<<[point,edge1.vertices[0],edge1]
                                points_curve<<edge2
                                # break
                            elsif edge1.curve && (edge1.vertices[1]==edge1.curve.vertices[0]) &&
                                point_in_polygon(edge1.vertices[1].position,[edge1.vertices[0].position,point])
                                points_edge1<<[point,edge1.vertices[1],edge1]
                                points_curve<<edge2
                                # break
                            elsif
                                edge1.curve && (edge1.vertices[0]==edge1.curve.vertices[-1]) &&
                                point_in_polygon(edge1.vertices[0].position,[edge1.vertices[1].position,point])
                                points_edge1<<[point,edge1.vertices[0],edge1]
                                points_curve<<edge2
                                # break
                            elsif edge1.curve && (edge1.vertices[1]==edge1.curve.vertices[-1]) &&
                                point_in_polygon(edge1.vertices[1].position,[edge1.vertices[0].position,point])
                                points_edge1<<[point,edge1.vertices[1],edge1]
                                points_curve<<edge2
                                # break
                            elsif point2.position.distance(point)>0.1 &&
                                !edge1.curve && point_in_polygon(edge1.vertices[1].position,[edge1.vertices[0].position,point])
                                points_edge1<<[point,edge1.vertices[1],edge1]
                                points_curve<<edge2
                            elsif point2.position.distance(point)>0.1 &&
                                !edge1.curve && point_in_polygon(edge1.vertices[0].position,[edge1.vertices[1].position,point])
                                points_edge1<<[point,edge1.vertices[0],edge1]
                                points_curve<<edge2

                            end
                        end
                    end
                else
                    point= Geom::closest_points(edge1.line,edge2.line)
                    if point
                        if point_in_polygon(point[1],edge2.vertices.collect{|e| e.position})
                            closest_points<<[point[0],point[1],edge1]
                            point2=edge1.vertices.min_by{|v| v.position.distance(point[0])}
                            if point2 && point2.position.distance(point[0])>0.0001
                                points_edge1<<[point[0],point2,edge1]
                                points_curve<<edge2
                            end

                        end
                    end
                end
            }
            return [points_curve,points_edge1,closest_points]
        end
        def self.intersec_arc_edges(arc,edges,out=false)
            points=[]
            points1=[]
            edges_arc,center,radius=get_edges_circle arc
            reverse=false
            if edges_arc[0].distance(arc.vertices[0].position)>0.01
                edges_arc.reverse!
                reverse=true
            end
            edges1_arc=edges_arc.each_cons(2).to_a
            edges1_arc+=arc.edges.collect{|e| [e.vertices[0].position,e.vertices[1].position]}
            edges1_arc.each{|e|
                edges.each{|edge1|
                    if edge1 && (center.distance_to_line(edge1.line)<radius)
                        e=e.collect{|i| Geom::Point3d.new(i)}
                        line=[e[0],(e[1]-e[0])]
                        if line[1].length!=0
                            p=Geom::intersect_line_line(edge1.line,line)
                            if  (p && out && point_in_polygon(p,e)) ||
                                (p && !out && point_in_polygon(p,edge1.vertices.collect{|v|v.position}) &&
                                point_in_polygon(p,e))
                                points<<edge1
                                break #if !out
                            end
                        end
                    end
                }
                #break if (!out && points.length>0) || (out &&  points.length>1)
            }
            edges1_arc=edges_arc.reverse.each_cons(2).to_a
            edges1_arc+=arc.edges.reverse.collect{|e| [e.vertices[0].position,e.vertices[1].position]}
            edges1_arc.each{|e|
                edges.each{|edge1|
                    if edge1 && (center.distance_to_line(edge1.line)<radius)
                        e=e.collect{|i| Geom::Point3d.new(i)}
                        line=[e[0],(e[1]-e[0])]
                        if line[1].length!=0
                            p=Geom::intersect_line_line(edge1.line,line)
                            if (p && !out && point_in_polygon(p,edge1.vertices.collect{|v|v.position}) &&
                                point_in_polygon(p,e))
                                points1<<edge1
                                break
                            end
                        end
                    end
                }
                #break  if (!out && points1.length>0) || (out &&  points1.length>1)
            } if !out
            ps=[points.uniq,points1.uniq]
            #ps.reverse! if reverse
            return ps
        end
        def self.get_length_points(points)
            length=0
            return length if points.length<2
            points.each_cons(2).to_a.each{|ps| length+=ps[0].distance(ps[1])}
            return length
        end
        def self.connect_extend()
            model=Sketchup.active_model
            sel=model.selection.to_a
            no_selected=model.active_entities.grep(Sketchup::Edge)-sel
            curvers=[]
            edges=[]
            edges1=[]
            edges_no=[]
            edge1=nil
            edges2=[]
            curver=[]
            sel.each{|e|
                if e && e.valid? && e.is_a?(Sketchup::Edge) && e.length!=0
                    if e.curve
                        curvers << e.curve
                    else
                        if e.vertices.any?{|v| (v.edges & sel).length<=1}
                            edges << e
                        elsif e.vertices.any?{|v| (v.edges & sel).length<=2}
                            edges1 << e
                        else
                            edges_no << e
                        end
                    end
                end
            }
            curvers.uniq!
            arcs=[]
            arcs=curvers.select{|c| self.curve_is_arc(c)}
            curver=curvers-arcs
            edges=edges1 if edges.empty?
            edges=edges_no if edges.empty?

            if edges.empty? && curver.length>0 && arcs.empty?
                if curver.length==1
                    edges<< curver[0].edges[0]
                    edges2<< curver[0].edges[-1]
                else
                    vers=curver.collect{|c| [c.vertices[0],c.vertices[-1]]}.flatten
                    vers.sort_by!{|v| v.edges.length}
                    ed=(vers[0].edges & sel)
                    cur=ed.collect{|e| e.curve if e.curve}.compact
                    cur.each{|c|
                        edges<<c.edges[0]
                        edges<<c.edges[-1]
                    }
                    curver=curver-cur
                    curver.each{|c| edges2+=c.edges}
                end
            end
            if edges.length>0 && arcs.empty?
                if curver.length>0 && edges2.empty?
                    curver.each{|c|
                        edges.each{|e|
                            ed1=get_intersect_curve_line(c,e)[0]
                            edges2+=ed1
                        }
                        edges2<< c.edges[0]
                        edges2<< c.edges[-1]
                    }
                end
                if edges2.empty? && edges.length>1
                    edges2=edges[1..-1]
                    edges=[edges[0]]
                end
            end
            edges2.uniq!
            if edges && !edges.empty? && edges2 && !edges2.empty?
                points_edge1=[]
                points_edge2=[]
                closest_points=[]
                edges.each{|edge1|
                    edges2.each{|edge2|
                        if edge1!=edge2
                            points_1,points_2=get_intersect_line_line(edge1,edge2)
                            points_edge1+=points_1
                            points_edge2+=points_2
                        end
                    }
                }
                if points_edge1.empty? && points_edge2.empty?
                    edges.each{|edge1|
                        edges2.each{|edge2|
                            if edge1!=edge2
                                points_1,points_2,closest=get_intersect_line_line(edge1,edge2,false)
                                points_edge1+=points_1
                                points_edge2+=points_2
                                closest_points+=closest
                            end
                        }
                    }
                end
                if (points_edge1+points_edge2).length>0
                    ps1=[]
                    if !points_edge1.empty?
                        ps=points_edge1#.select{|p| p[0].distance(p[1].position)>0.0001}
                        ps=ps.chunk{|p| (p[1].edges & sel).length }.sort_by{|p| p[0]}
                        ps=ps[0][1].sort_by{|p| p[0].distance p[1].position}
                        ps=ps.chunk{|p| (p[1].edges).length }.sort_by{|p| p[0]}
                        ps=ps[0][1].sort_by{|p| p[0].distance p[1].position}
                        if curver.length>0 && ps.length>0
                            ps1 += [ps[0]] + points_edge2.select{|p| p[0].distance(ps[0][0]) <= 0.0001}
                        else
                            ps1+=ps
                        end
                    end
                    if !points_edge2.empty?
                        ps=points_edge2#.select{|p| p[0].distance(p[1].position)>0.0001}
                        ps=ps.chunk{|p|(p[1].edges & sel).length }.sort_by{|p| p[0]}
                        ps=ps[0][1].sort_by{|p| p[0].distance p[1].position }
                        ps=ps.chunk{|p|(p[1].edges ).length }.sort_by{|p| p[0]}
                        ps=ps[0][1].sort_by{|p| p[0].distance p[1].position }
                        if curver.length>0 && ps.length>0
                            ps1 += [ps[0]] + points_edge1.select{|p| p[0].distance(ps[0][0]) <= 0.0001}
                        else
                            ps1+=ps
                        end
                    end
                    ps1.uniq{|p| p[1]}
                    ps1_curve=ps1.select{|p| p[2].curve }
                    if ps1_curve.length>0
                        ps1_curve.sort_by!{|p| p[0].distance(p[1].position)}
                        ps1=[ps1_curve[0]] + ps1.select{|p| p[0].distance(ps1_curve[0][0])<0.001}
                    end
                    Sketchup.active_model.start_operation Strings["Connect Extend"],false,false,false
                    begin
                    g=model.active_entities.add_group
                    self.add_edges_extend(g,ps1,closest_points)
                    g.explode
                    Sketchup.active_model.commit_operation
                    rescue=>error
                        p error,error.backtrace.join("\n")
                        Sketchup.active_model.abort_operation
                    end
                    model.selection.add(model.active_entities.grep(Sketchup::Edge) - no_selected)
                end
            elsif arcs.length>0
                if arcs.length==1 && curver.length==1
                    points=intersec_arc_edges(arcs[0],curver[0].edges)
                    if points.flatten.length>1
                        edges=points.flatten
                    else
                        points=intersec_arc_edges(arcs[0],[curver[0].edges[0],curver[0].edges[-1]],true)
                        if points.flatten.length>0
                            edges=points.flatten
                        else
                            arcs=[]
                        end
                    end
                end
                if edges.empty?
                    if arcs.length>1
                        if arcs.length==2
                            points1=intersec_arc_edges(arcs[0],arcs[1].edges)
                            points2=intersec_arc_edges(arcs[1],arcs[0].edges)
                            if (points1.flatten.length>points2.flatten.length)
                                edges+=points1.flatten
                                arcs.delete_at(1)
                            elsif(points1.flatten.length<points2.flatten.length)
                                edges+=points2.flatten
                                arcs.delete_at(0)
                            elsif (points1.flatten.length==points2.flatten.length) &&
                                points1.flatten.length>0
                                edges+=points1.flatten+points2.flatten
                            else
                                edges=[nil]
                            end
                        else
                            edges=[nil]
                        end
                    else
                        edges=[nil]
                    end
                end
                points_edge1=[]
                points_edge2=[]
                points_edge3=[]
                closest_points=[]
                curves=[]
                arcs.each{|c|
                    edges.each_index{|i|
                        edge1=edges[i]
                        if edge1 && ( !edge1.curve || (edge1.curve && !curve_is_arc(edge1.curve)))
                            points1,points2,closest=get_intersect_curve_line(c,edge1,true,edges1)
                            points_edge1+=points2
                            closest_points+=closest
                            points1.each{|e|
                                points1,points2,closest=get_intersect_line_line(edge1,e,true)
                                points_edge1+=points1
                                #points_edge2+=points2
                                closest_points+=closest
                            }
                        end
                    }
                }
                if arcs.length<=2
                    radius=nil
                    arcs.each{|c|
                        edges.each_index{|i|
                            edge1=edges[i]
                            edges_arc,center,radius=get_edges_circle c
                            edges_arc.reverse!
                            if edge1 && (center.distance_to_line(edge1.line)<radius)
                                points_edge=edge1.vertices.collect{|p| p.position}
                                ps=[]
                                ps1=[]
                                edges1_arc=edges_arc.each_cons(2).to_a
                                edges1_arc.each{|e|
                                    e=e.collect{|i| Geom::Point3d.new(i)}
                                    ps<<e[0]
                                    line=[e[0],(e[1]-e[0])]
                                    if line[1].length!=0
                                        p=Geom::intersect_line_line(edge1.line,line)
                                        if p && point_in_polygon(p,e) &&
                                            (!edge1.curve || (edge1.curve && point_in_polygon(p,points_edge)))
                                            ps << p if p.distance(ps[-1])>0
                                            if p.distance(edge1.start.position) <= p.distance(edge1.end.position)
                                                points_edge2<<[p,edge1.start,edge1]
                                            else
                                                points_edge2<<[p,edge1.end,edge1]
                                            end
                                            break #if ps.length>1
                                        end
                                    end
                                }
                                edges1_arc.reverse.each{|e|
                                    e=e.collect{|i| Geom::Point3d.new(i)}
                                    ps1<<e[-1]
                                    line=[e[1],(e[0]-e[1])]
                                    if line[1].length!=0
                                        p=Geom::intersect_line_line(edge1.line,line)
                                        if p && point_in_polygon(p,e) &&
                                            (!edge1.curve || (edge1.curve && point_in_polygon(p,points_edge)))
                                            ps1 << p if p.distance(ps1[-1])>0
                                            if p.distance(edge1.start.position) <= p.distance(edge1.end.position)
                                                points_edge2<<[p,edge1.start,edge1]
                                            else
                                                points_edge2<<[p,edge1.end,edge1]
                                            end
                                            break #if ps1.length>1
                                        end
                                    end
                                }
                                cur=[ps,ps1]
                                curves<<[cur,c]
                            elsif !edge1
                                as=arcs.select{|a| a!=c}
                                edges_arc1=edges_arc.each_cons(2).to_a
                                as.each{|a|
                                    edges_arc,center,radius=get_edges_circle a
                                    edges_arc.reverse!
                                    edges_arc1.each{|edge1|
                                        points_edge=edge1
                                        ps=[]
                                        ps1=[]
                                        edges1_arc=edges_arc.each_cons(2).to_a
                                        edges1_arc.each{|e|
                                            e=e.collect{|i| Geom::Point3d.new(i)}
                                            ps<<e[0]
                                            line=[e[0],(e[1]-e[0])]
                                            edge1_line=[edge1[0],(edge1[1]-edge1[0])]
                                            if line[1].length!=0 && edge1_line[1].length!=0
                                                p=Geom::intersect_line_line(edge1_line,line)
                                                if p && point_in_polygon(p,e) &&
                                                    (point_in_polygon(p,points_edge))
                                                    ps << p if p.distance(ps[-1])>0
                                                    if p.distance(edge1[0]) <= p.distance(edge1[1])
                                                        points_edge3<<[p,edge1[0],edge1]
                                                    else
                                                        points_edge3<<[p,edge1[1],edge1]
                                                    end
                                                    break #if ps.length>1
                                                end
                                            end
                                        }
                                        edges1_arc.reverse.each{|e|
                                            e=e.collect{|i| Geom::Point3d.new(i)}
                                            ps1<<e[-1]
                                            line=[e[1],(e[0]-e[1])]
                                            edge1_line=[edge1[0],(edge1[1]-edge1[0])]
                                            if line[1].length!=0 && edge1_line[1].length!=0
                                                p=Geom::intersect_line_line(edge1_line,line)
                                                if p && point_in_polygon(p,e) &&
                                                    (point_in_polygon(p,points_edge))
                                                    ps1 << p if p.distance(ps1[-1])>0
                                                    if p.distance(edge1[0]) <= p.distance(edge1[1])
                                                        points_edge3<<[p,edge1[0],edge1]
                                                    else
                                                        points_edge3<<[p,edge1[1],edge1]
                                                    end
                                                    break #if ps1.length>1
                                                end
                                            end
                                        }
                                        if points_edge3.length>0
                                            cur=[ps,ps1]
                                            curves<<[cur,c]
                                        end
                                    }
                                }
                                if points_edge3.empty? && edges_arc && edges_arc.length>1
                                    curves<<[[edges_arc,edges_arc.reverse],c]
                                end
                            end
                        }
                    }
                    if arcs.length==1 && radius
                        puts ("Radius = " + radius.to_l.to_s)
                    end
                else
                    arcs.each{|c|
                        edges_arc,center,radius=get_edges_circle c
                        if edges_arc && edges_arc.length>1
                            curves<<[[edges_arc,edges_arc],c]
                        end
                    }
                end
                ps1=[]
                if !points_edge1.empty?
                    ps=points_edge1#.select{|p| p[0].distance(p[1].position)>0.0001 }
                    ps=ps.chunk{|p| (p[1].edges & sel).length }.sort_by{|p| p[0]}
                    if ps.length>0
                        ps=ps[0][1]
                        ps=ps.chunk{|p| (p[1].edges).length }.sort_by{|p| p[0]}
                        ps=ps[0][1]
                        ps=ps.sort_by{|p| p[0].distance p[1].position}
                        if ps.length>0
                            ps1 += [ps[0]]
                        end
                    end
                end
                # if points_edge2.length>0
                    # ps=points_edge2#.select{|p| p[0].distance(p[1].position)>0.0001}
                    # ps=(ps.chunk{|p| (p[1].edges & sel).length}).sort_by{|p| p[0]}
                    # if ps.length>0
                        # ps=ps[0][1]
                        # ps=(ps.chunk{|p| (p[1].edges).length}).sort_by{|p| p[0]}
                        # ps=ps[0][1]
                        # ps=ps.sort_by{|p| p[0].distance p[1].position }
                        # if ps.length>0
                            # ps1 += [ps[0]]
                        # end
                    # end
                # end
                ps1.uniq{|p| p[1]}
                Sketchup.active_model.start_operation Strings["Connect Extend"],true,false,false
                begin
                g=model.active_entities.add_group
                points_edge1.uniq!{|e| e[1]}
                points_edge2.uniq!{|e| e[1]}
                points_edges= (points_edge1+points_edge2).uniq{|e| e[1]}
                ps2=[]
                curves_end=[]
                curs=curves.chunk{|c| c[1]}.to_a
                curves_end=[]
                curs.each{|c1|
                    p1 = []
                    p2 = []
                    c2=c1[1]
                    if ((c1[0].vertices[0].edges & sel).length) > ((c1[0].vertices[-1].edges & sel).length)
                        i=1
                        ps=(c2.select{|c| c[0][i].length>0}).sort_by{|c| get_length_points(c[0][i])}
                        if ps[0] && ps[0][0] && ps[0][0][i] && ps[0][0][i].length>0
                            p=ps[0][0][i]
                            curves_end<<[p,c1[0]] if p.length>0
                        end
                    elsif (c1[0].vertices[0].edges & sel).length <(c1[0].vertices[-1].edges & sel).length
                        i=0
                        ps=(c2.select{|c| c[0][i].length>0}).sort_by{|c| get_length_points(c[0][i])}
                        if ps[0] && ps[0][0] && ps[0][0][i] && ps[0][0][i].length>0
                            p=ps[0][0][i]
                            curves_end<<[p,c1[0]] if p.length>0
                        end
                    elsif (c1[0].vertices[0].edges).length > (c1[0].vertices[-1].edges).length
                        i=1
                        ps=(c2.select{|c| c[0][i].length>0}).sort_by{|c| get_length_points(c[0][i])}
                        if ps[0] && ps[0][0] && ps[0][0][i] && ps[0][0][i].length>0
                            p=ps[0][0][i]
                            curves_end<<[p,c1[0]] if p.length>0
                        end
                    elsif (c1[0].vertices[0].edges).length <(c1[0].vertices[-1].edges).length
                        i=0
                        ps=(c2.select{|c| c[0][i].length>0}).sort_by{|c| get_length_points(c[0][i])}
                        if ps[0] && ps[0][0] && ps[0][0][i] && ps[0][0][i].length>0
                            p=ps[0][0][i]
                            curves_end<<[p,c1[0]] if p.length>0
                        end
                    else
                        i=1
                        ps=(c2.select{|c| c[0][i].length>0}).sort_by{|c| get_length_points(c[0][i])}
                        if ps[0] && ps[0][0] && ps[0][0][i] && ps[0][0][i].length>0
                            p=ps[0][0][i]
                            curves_end<<[p,c1[0]] if p.length>0
                        end
                        i=0
                        ps=(c2.select{|c| c[0][i].length>0}).sort_by{|c| get_length_points(c[0][i])}
                        if ps[0] && ps[0][0] && ps[0][0][i] && ps[0][0][i].length>0
                            p=ps[0][0][i]
                            curves_end<<[p,c1[0]] if p.length>0
                        end
                    end

                }
                curves_end.sort_by!{|v| get_length_points(v[0])}
                curs=curves_end.find_all{|c| c && c[0].length>1 && get_length_points(c[0])>0}
                if (arcs.length==1)
                    curs=[curs[0]]
                elsif arcs.length==2
                    curs_t=curs[1..-1].select{|c| c[1]!=curs[0][1]}
                    curs_t=curs_t.sort_by{|v| get_length_points(v[0])}.find_all{|c| c && c[0].length>1 && c[0][-1]==curs[0][0][-1]}
                    curs=[curs[0],curs_t[0]]
                end
                curs.each{|cur|
                    if cur && cur[0].length>1
                        ed=g.entities.add_curve cur[0]
                        if ed
                            e1=cur[1].edges[0]
                            ed.each{|e|
                                if e && e1 && e1.is_a?(Sketchup::Edge)
                                    e.layer=e1.layer
                                    e.material=e1.material
                                    e.hidden=true if e1.hidden?
                                    e.soft=true if e1.soft?
                                    e.smooth=true if e1.smooth?
                                end
                            }
                        end
                        points_edges.each{|p|
                            if p[0].distance(cur[0][-1])<0.001 || p[0].distance(cur[0][0])<0.001
                                ps2 << p
                            end
                        }
                    end
                } if points_edge1.empty? || points_edge1.any?{|e| (e[1].edges & sel).length>1}
                ps2=ps1 if ps2.empty?
                self.add_edges_extend(g,ps2,closest_points)
                g.explode
                Sketchup.active_model.commit_operation
                rescue=>error
                    p error,error.backtrace.join("\n")
                    Sketchup.active_model.abort_operation
                end
                model.selection.add(model.active_entities.grep(Sketchup::Edge) - no_selected)
            end
        end
        def self.main(mode=0)
            if mode==0
                self.connect
            elsif mode==1
                self.connect_extend
            end
        end
    end
end